
""" This is the main program of Z-curve Plotter (3D) """
import numpy as np
import pkg_resources
import json
import sys
import argparse
from Bio import SeqIO
from ZcurvePy import __version__
from ZcurvePy.Util import download_acc, has_gui
import matplotlib
HAS_GUI = has_gui()
matplotlib.use("TkAgg" if HAS_GUI else "Agg")
from matplotlib import pyplot as plt
from operator import methodcaller
from _ZcurvePy import ZcurvePlotter

# curve names
names = {
    "RY" : "RY disparity",
    "MK" : "MK disparity",
    "WS" : "WS disparity",
    "AT" : "AT disparity",
    "GC" : "GC disparity",
    "XP" : "x' curve",
    "YP" : "y' curve",
    "ZP" : "z' curve",
    "AP" : "AT' curve",
    "GP" : "GC' curve",
    "CG" : "CpG' curve",
}

prog_infor = \
f"""
Z-curve Plotter -- A program to plot visual Z-curves

Authors:    Zhang Zetong, Gao Feng
Version:    {__version__}
Contact:    fgao@tju.edu.cn
Copyright:  Bioinformatic Center, Tianjin University, China
"""

epilog_infor = \
"""
The content of setting file should be like:

{
    "plotter": [
        {   // The first sequence
            "start": 1000,         // Plotting start point (index starts from 0)
            "stop": 30000,         // Plotting stop point (index starts from 0)
            "comp": false,         // Plot reverse complement
            "window": 100,         // Mean smoothing window
            "intv": 10,            // Sampling interval
            "curve2d": "RY,MK",    // 2D-curves type
            "curve3d": "RY:MK:ZP"  // 3D-curve types
        }
        {   // The second sequence
            "start": 500,
            "stop": 40000,
            "comp": true,
            "window": 100,
            "intv": 5,
            "curve2d": "XP,YP,ZP"
        },
        // ...
    ]
}

The valid curve type codes and their names are listed below:

"""

for index, (key, value) in enumerate(names.items()):
    epilog_infor += f"({index + 1})\t{key} - {value}\n"

epilog_infor += \
"""

Note: 1. If any of the param is not existing, the programs will use defaults.
      2. When the 'stop' value is larger than 'start' value, the program
         will default the topology property of the sequence to ring and go on.
      3. 'intv' only effects the sampling interval when plotting visual curves.

Tips: Use the Z-curve method more flexibly through our ZcurvePy's APIs! 
"""

arg_fasta = "    input genome files as FASTA format (*.fa; *.fasta; *.fna)"
arg_gbk   = "    input genome files as GenBank format (*.gb; *.genbank; *.gbk; *.gbff)"
arg_acc   = "    input NCBI accession number (comma-splited; *.txt)"
arg_set   = "    external setting file as JSON format (*.json)"
arg_out   = "    output data file of all curves as JSON (*.json)"
arg_png   = "    output graphic file as PNG picture (*.png)"
arg_show  = "    show graphic user interface or not (default: False)"

# curve types
types = {
    "RY"  : "RY_disparity",
    "MK"  : "MK_disparity",
    "WS"  : "WS_disparity",
    "AT"  : "AT_disparity",
    "GC"  : "GC_disparity",
    "XP"  : "x_prime_curve",
    "YP"  : "y_prime_curve",
    "ZP"  : "z_prime_curve",
    "AP"  : "AT_prime_curve",
    "GP"  : "GC_prime_curve",
    "CG"  : "CpG_prime_curve"
}

def checkInputs(args: dict, command_name: str):
    records, fasta, genbank, acc = [], args['fasta'], args['genbank'], args['accession']

    try:
        # read fastas
        if fasta is not None:
            single_fastas = fasta.split(',')
            for single_fasta in single_fastas:
                records += list(SeqIO.parse(single_fasta, "fasta"))
        
        # read local or downloaded genbanks
        if genbank is not None:
            single_genbanks = genbank.split(',')
            for single_genbank in single_genbanks:
                records += list(SeqIO.parse(single_genbank, "genbank"))

        # read download files
        if acc is not None:
            downloads = download_acc(acc, command_name)
            for download in downloads:
                records += list(SeqIO.parse(download, "fasta"))
    except Exception as exception:
        print(f"Biopython: [error] failed to read input files! ")
        print(exception)
        sys.exit()

    num_records = len(records)
    
    if num_records == 0:
        print(f"{command_name}: [error] no input files was found!")
        sys.exit()
    else:
        print(f"{command_name}: [infor] detected {num_records} sequences. ")
    
    return records


def handleInputs3D(
    records: list, 
    args: dict, 
    command_name: str,
):
    with open(args['settings'], 'r', encoding='utf-8') as file:
        settings = json.load(file)['plotter']
    num_settings, curve_list = len(settings), []

    for i, record in enumerate(records):
        if i < num_settings:
            setting = settings[i]

            start, stop = 0, len(record)
            if 'start' in setting.keys():
                start = int(setting['start'])
            if 'stop' in setting.keys():
                stop = int(setting['stop'])
            
            if start > stop:
                record = record[start:] + record[:stop]
            elif start < stop:
                record = record[start:stop]
            else:
                print(f"{command_name}: [error] The slice length of {record.id} is zero!")
                sys.exit()

            if "comp" in setting.keys() and setting['comp']:
                record = record.reverse_complement()
            
            intv = int(setting['intv'] if "intv" in setting.keys() 
                       else max(len(record)/ 1E7, 1))
            curve_types = ['RY', 'MK', 'WS']

            if "curve3d" in setting.keys():
                curve_types = setting['curve3d'].split(":")
                if len(curve_types) != 3:
                    print(f"{command_name}: [error] The number of componets must be 3 in 3D-mode!")
                    sys.exit()
            else:
                print(f"{command_name}: [error] No valid settings found for 3D-mode !")
                sys.exit()
            
            curves = []
            plotter = ZcurvePlotter(record)

            for typ in curve_types:
                if typ not in types.keys():
                    print(f"{command_name}: [error] Invalid curve type {typ}")
                    sys.exit()
                window = int(setting['window']) if 'window' in setting.keys() else 0
                retrs = methodcaller(types[typ], window=window, return_n=False)(plotter)
                yvalues = retrs[0] if isinstance(retrs, tuple) else retrs
                curves.append(np.array(yvalues))
                
            curve_list.append((record.id, curves, intv))
    
    return curve_list
        

def visualize3D(curve_list: list, png: str, show: bool):
    plt.rc('font', family='Times New Roman')
    plt.rcParams.update({"font.size": 14})
    fig = plt.figure(figsize=(10,6))

    ax = fig.add_subplot(projection='3d')
    ax.set_title('Z-curve')

    for item in curve_list:
        label, (x, y, z), intv = item
        ax.plot(x[::intv], y[::intv], z[::intv], label=label)
        
    ax.set_xlabel("X Axis", labelpad=10)
    ax.set_ylabel("Y Axis", labelpad=10)
    ax.set_zlabel("Z Axis", labelpad=10)
    ax.legend()
    
    if png is not None:
        fig.savefig(png)
    if show and HAS_GUI:
        try:
            manager = plt.get_current_fig_manager()
            fig.canvas.manager.set_window_title("Z-curve Plotter (Powered by Matplotlib)")
            res_path = pkg_resources.resource_filename("ZcurvePy", "assets/image/icon.ico")
            manager.window.iconbitmap(res_path)
            plt.show()
        except Exception as exception:
            print("zcurve-plotter-3d: [error] unable to create GUI! ")
            print(exception)


def main():
    command_name = 'zcurve-plotter-3d'

    parser = argparse.ArgumentParser(
        prog=command_name,
        description=prog_infor,
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=epilog_infor
    )

    parser.add_argument('-f', '--fasta', metavar='\b', required=False, type=str, help=arg_fasta)
    parser.add_argument('-g', '--genbank', metavar='\b', required=False, type=str, help=arg_gbk)
    parser.add_argument('-a', '--accession', metavar='\b', required=False, type=str, help=arg_acc)
    parser.add_argument('-s', '--settings', metavar='\b', required=True, type=str, help=arg_set)
    parser.add_argument('-o', "--output", metavar='\b', required=False, default=None, type=str, help=arg_out)
    parser.add_argument('-p', "--png", metavar='\b', required=False, type=str, default=None, help=arg_show)
    parser.add_argument('-v', "--show", metavar='\b', required=False, type=bool, default=True, help=arg_show)

    args = vars(parser.parse_args())

    records = checkInputs(args, command_name)
    curve_list = handleInputs3D(records, args, command_name)

    png, show = args['png'], args['show']
    if png is not None or show:
        visualize3D(curve_list, png, show)
    
    output = args['output']
    if output is not None:
        contents = []
        for curve in curve_list:
            name, (x, y, z), _ = curve
            contents.append({name: {
                'x': np.round(x, decimals=2).tolist(), 
                'y': np.round(y, decimals=2).tolist(), 
                'z': np.round(z, decimals=2).tolist()
            }})
        to_save = json.dumps(contents)
        with open(output, "w") as file:
            file.write(to_save)



from ZcurvePy import __version__
f"""
    This is the util module of ZcurvePy

    We are happy to see that the APIs provided here are also
    useful for your project, so you can just use it in your
    Python scripts.

    Authors:    Zetong Zhang, Feng Gao
    Version:    { __version__ }
    Copyright:  Copyright 2025 TUBIC
    License:    MIT License
    Date:       2025-02-26
"""
import os
import sys
from Bio.SeqIO import parse, read, write
from Bio.SeqRecord import SeqRecord
from Bio import Entrez

Entrez.email = "flypigaround@163.com"
Entrez.api_key = "24a9ecda42217cd6d8f24058a4d6df456c09"

# where the downloaded files saved
CACHE_PATH = "ncbi_cache"

def download_acc(acc, command_name=__name__, form='fasta'):
    """
    Download genome files as genbank from NCBI database
    by accession using ncbi-acc-download

    Args:
        acc (str):          comma-splited accesions (or a
                            text files stores accessions)
        command_name (str): the name of command. If the
                            script is not called by console
                            commands, it is set as the name
                            of the module defaultly. 
        form (str):         the format of to-download files
    Returns:
        ready (list):       The save paths of the files 
                            downloaded and are ready to use.
    """
    to_download, ready = None, []

    if not os.path.exists(CACHE_PATH):
        os.mkdir(CACHE_PATH)
    
    if acc.endswith(".txt"):
        file = open(acc, 'r')
        to_download = file.readlines()
        file.close()
    else:
        to_download = acc.split(',')
    
    for item in to_download:
        item = item.strip()

        if len(item) == 0:
            continue
            
        file_name = item + ('.fa' if form == 'fasta' else '.gbk')
        save_path = os.path.join(CACHE_PATH, file_name)

        if not os.path.exists(save_path):
            try:
                handle = Entrez.efetch(
                    db="nucleotide",
                    id=item,
                    rettype=form,
                    retmode="text")
                record = read(handle, form)
                handle.close()
                write(record, save_path, form)
            except Exception as e:
                print(f"{command_name}:{e}")
                sys.exit(0)

        ready.append(save_path)
    return ready

def extract_CDS(genome):
    """
    Extract all the CDSs sequence from annotated genome files.
    Usually a GenBank file.

    Args:
        genome (str):   path to the genome file
    
    Returns:
        CDSs (list):    a list of CDS records
    """
    CDSs, records = [], parse(genome, "genbank")

    for record in records:
        for feature in record.features:
            if feature.type == "CDS":
                name = feature.qualifiers['locus_tag'][0]
                CDS_seq = feature.extract(record.seq)
                CDS_record = SeqRecord(CDS_seq, id=name)
                CDSs.append(CDS_record)
    
    return CDSs


def has_gui():
    """
    Check whether the current system has GUI

    Returns:
        bool: whether the current system has GUI
    """
    # Linux
    if sys.platform.startswith('linux'):
        if not os.environ.get('DISPLAY'):
            return False
        try:
            import tkinter as tk
            root = tk.Tk()
            root.withdraw()
            root.update()
            root.destroy()
            return True
        except Exception:
            return False
    # Windows
    elif sys.platform == 'win32':
        try:
            import tkinter as tk
            tk.Tk().withdraw()
            return True
        except Exception:
            try:
                from ctypes import windll
                return windll.user32.GetSystemMetrics(0) != 0
            except Exception:
                return False
    # Unsupported OS
    else:
        return False



__version__ = '1.6.0'
__upgrade__ = '2025-04-29'
f"""
    ZcurvePy -- High performance Python toolkit for the Z-curve theory!

    This is the Python package init module of ZcurvePy v{__version__} released.

    All the C/C++ apis can be called through this module. If you don't
    want to import additional third-party module (such as modules from
    scikit-learn) or APIs written in pure Python provided by ZcurvePy,
    (such as ZcurveBuilder), use 'import _ZcurvePy' instead of 'import
    ZcurvePy).

    Authors:    Zetong Zhang, Feng Gao
    Copyright:  Copyright 2025 TUBIC
    License:    MIT License
    Date:       {__upgrade__}

    Appendix Table:
            Correspondence between bases and chars in this software
    | char | base       | char | base       | char | base    | char | base    |
    |------|------------|------|------------|------|---------|------|---------|
    |  A   | A          |  B   | C, G, T    |  C   | C       |  D   | A, G, T |
    |  E   | null       |  F   | null       |  G   | G       |  H   | A, C, T |
    |  I   | A, G, C, T |  J   | null       |  K   | G, T    |  L   | null    |
    |  M   | A, C       |  N   | A, G, C, T |  O   | null    |  P   | null    |
    |  Q   | null       |  R   | A, G       |  S   | G, C    |  T   | T       |
    |  U   | T          |  V   | A, G, C    |  W   | A, T    |  X   | null    |
    |  Y   | C, T       |  Z   | A          |

    [Remarks]:
    1.  'null' means the software will just skip the character when read strings;
    2.  Degenerate symbols will be handled using rule of frequency, for example:
        'M' will be regarded as 50% A + 50% C and vectorized as [0.5, 0, 0.5, 0]
    3.  'I' means hypoxanthine and may be paired with any type of bases; 'Z' means
        diaminopurine and can only be paired with 'A'. (Zhou Y, et al. Science, 
        2021, 372(6541): 512-516.)
"""
import sys
# Up regulate max recursion for ZcurveSegmenter
sys.setrecursionlimit(10000)
from operator import methodcaller
from copy import deepcopy
# Scikit-learn APIs
from sklearn.svm import SVC
from sklearn.preprocessing import StandardScaler, MinMaxScaler
# Numpy Support
import numpy as np

# ZcurvePy C/C++ APIs (Classes)
from _ZcurvePy import ZcurveEncoder
from _ZcurvePy import ZcurvePlotter
# ZcurvePy C/C++ APIs (Multi-thread)
from _ZcurvePy import BatchZcurveEncoder
from _ZcurvePy import BatchZcurvePlotter
# ZcurvePy C/C++ APIs (Method)
from _ZcurvePy import shuffle
from _ZcurvePy import decode

class ZcurveBuilder:
    """
    A simple API for building gene recognizer. It includes basic modules such 
    as feature extraction, preprocessing, negative sampling and model training.
    Each module is customizable. Core algorithm of ZCURVE 2.0 / 3.0.

    Args:
        encoder (Callable):  
            A function or object that can be called with a list of supported 
            sequences and returns a feature matrix of the data set. Default 
            is a BatchZcurveEncoder object that yields 189-dim features.

        model (object):  
            A machine learning model object that have `fit` and `predict_proba`
            method. Default is a RBF-SVM model from `sklearn.svm.SVC` with 
            balanced class weights.

        standard (bool):  
            Whether to apply standardization to the data. (Use `sklearn.
            preprocessing.StandardScaler`)  
            Default: True

        normal (bool):  
            Whether to apply normalization to the data. (Use `sklearn.
            preprocessing.MinMaxScaler`)  
            Default: False  
        
        neg_pos_ratio (int):  
            The ratio between negative samples and positive samples, decides 
            how many negative sequences should be obtained from each positive 
            case sequence.  
            Parameter for `ZcurveBuilder.shuffle`, default is 5.

        seed (int):  
            Random seed.
        
        n_jobs (int):  
            Number of CPU cores used. All are used by default.
    """
    def __init__(
        self,
        encoder=None,
        model=None,
        standard=True,
        normal=False,
        neg_pos_ratio=5,
        seed=42,
        n_jobs=-1
    ):
        self.encoder = encoder
        if self.encoder is None:
            hyper_params = [
                {'n': 1, 'local': True},
                {'n': 2, 'local': True},
                {'n': 3, 'local': True}
            ]  # Final number of generated parameters: 189
            self.encoder = BatchZcurveEncoder(
                hyper_params=hyper_params,
                n_jobs=n_jobs
            )
        
        self.model = model
        if self.model is None:
            self.model = SVC(
                kernel='rbf', 
                class_weight='balanced',
                random_state=seed
            )
        
        self.scaler = StandardScaler() if standard is True else None
        self.minmax = MinMaxScaler() if normal is True else None

        self.neg_pos_ratio = neg_pos_ratio
        self.seed = seed
        self.n_jobs = n_jobs
    
    @staticmethod
    def has_method(obj, method_name):
        """ Determine whether an object has the member function. """
        return callable(getattr(obj, method_name, None))
    
    def fit(
        self, 
        pos_data,
    ):
        """
        Fit the model after negative samples are generated and the dataset is preprocessed.

        Args:
            pos_data (object):  
                Positive example sequence data sets generally use reference gene sequences 
                as positive examples.
        """
        gen_seed = deepcopy(pos_data)
        gen_data = shuffle(
            records=gen_seed,
            ratio=self.neg_pos_ratio,
            seed=self.seed,
            n_jobs=self.n_jobs
        )

        pos_features = np.array(self.encoder(pos_data))
        neg_features = np.array(self.encoder(gen_data))
        pos_labels = np.ones(pos_features.shape[0])
        neg_labels = np.zeros(neg_features.shape[0])

        X = np.concatenate((pos_features, neg_features), axis=0)
        y = np.concatenate((pos_labels, neg_labels), axis=0)

        if self.scaler is not None:
            X = self.scaler.fit_transform(X)
        if self.minmax is not None:
            X = self.minmax.fit_transform(X)

        self.model.fit(X, y)
    
    def predict(
        self,
        data,
        threshold=None
    ):
        """
        Predict whether a set of sequences is a gene coding sequence.

        Args:
            data (object):  
                Sequence dataset to predict.
            threshold (float):
                Label judgment threshold. Greater than the threshold 
                is a positive sample, and less than the threshold is 
                a negative sample.
        
        Returns:
            labels (ndarray): 
                Prediction label for the sample.
            scores (ndarray):
                The predicted score of the sample.
        """
        features = self.encoder(data)
        if self.scaler is not None:
            features = self.scaler.transform(features)
        if self.minmax is not None:
            features = self.minmax.transform(features)
        
        if self.has_method(self.model, "decision_function"):
            scores = self.model.decision_function(features)
            threshold = 0 if threshold is None else threshold
            labels = np.array(scores > 0, dtype='int')
        elif self.has_method(self.model, "predict_proba"):
            scores = self.model.predict_proba(features)
            threshold = 0.5 if threshold is None else threshold
            labels = np.array(scores > 0.5, dtype='int')
        else:
            raise Exception("Unsupported model types")
        
        return labels, scores


class ZcurveSegmenter:
    """
    Z-curve segmenter based on genome order number, which can be considered an 
    extended version of the GC-Profile's core. It has 7 modes and can be used 
    for edge recognition of genomic islands, CpG islands, AT-rich regions and 
    other structures.

    Args:
        mode (str): The mode of Z-curve segmenter. It can be one of the following:
            'GN': Using genome order number to segment Z-curves;
            'RY': Using RY order number to segment RY-profile;
            'MK': Using MK order number to segment MK-profile;
            'WS': Using WS order number to segment GC-profile;
            'AT': Using AT order number to segment AT-skew;
            'GC': Using GC order number to segment GC-skew;
            'CG': Using CpG order number to segment CpG-profile.
        halting (int): The halting value of segmenter. 
            Defaults to 100.
        min_len (int): The minimum length of fragments. 
            Defaults to 3000.
        max_depth (int): The maximum depth of iteration. 
            Defaults to 9999.
    """
    # Modes and their corresponding functions
    func = {
        'GN': 'genome_dS_curve',
        'RY': 'RY_dS_curve',
        'MK': 'MK_dS_curve',
        'WS': 'WS_dS_curve',
        'AT': 'AT_dS_curve',
        'GC': 'GC_dS_curve',
        'CG': 'CpG_dS_curve'
    }

    def __init__(
        self,
        mode,
        halting=100,
        min_len=3000,
        max_depth=9999
    ):
        self.mode = mode
        self.halting = halting
        self.min_len = min_len
        self.max_depth = max_depth
        self.executed = False
        self.seg_points = []
    
    def _recursion(
        self, 
        record, 
        offset: int, 
        depth: int
    ):
        """
        Recursive function for segmenting Z-curves.
        """
        plotter = ZcurvePlotter(record)
        max_point, max_value = methodcaller(self.func[self.mode], only_m=True)(plotter)
        left_len, right_len = max_point, len(record) - max_point

        if (max_value > self.halting and
            left_len >= self.min_len and right_len >= self.min_len):
            self.seg_points.append((max_point + offset, max_value))

            if depth < self.max_depth:
                # Segment left subsequence
                left_seq = record[:max_point]
                self._recursion(left_seq, offset, depth + 1)
                del left_seq
                # Segment right subsequence
                right_seq = record[max_point:]
                self._recursion(right_seq, offset + max_point, depth + 1)
                del right_seq
    
    def run(
        self, 
        record
    ):
        """
        Run the segmenter.

        Args:
            record (str): The sequence to be segmented.

        Returns:
            list:   A list of tuples, each tuple contains the 
                    position and the value of the segment point.
        """
        if not self.executed:
            self._recursion(record, 0, 0)
            self.seg_points.sort(key=lambda x : x[0])
            self.executed = True
            return self.seg_points
        else:
            raise Exception("Please reset the object first to segment another sequence!")
    
    def reset(self):
        """ 
        Reset the segmenter. 
        Clear segmentation points generated in the last run.
        """
        self.seg_points = []
        self.executed = False


""" This is the main program of Z-curve Encoder """
# PASS 2025-03-03
from _ZcurvePy import BatchZcurveEncoder
from ZcurvePy import __version__
from ZcurvePy.Util import download_acc, extract_CDS
from Bio.SeqIO import parse
import argparse
import json
import time
import sys

# program information
prog_infor = \
f"""
Z-curve Encoder -- A program to encode Z-curve parameters

Authors:    Zhang Zetong, Gao Feng
Version:    {__version__}
Contact:    fgao@tju.edu.cn
Copyright:  Bioinformatic Center, Tianjin University, China
"""

# appendix information
epilog_infor = \
"""
The content of setting file should be like:

{
    "encoder": {
        // Corresponds to 'hyper_params' of ZcurveEncoder.__init__
        "hyper_params": [
            {"k": 1, "phase": 3, "freq": true, "local": true},
            {"k": 2, "phase": 3, "freq": true, "local": true},
            {"k": 3, "phase": 3, "freq": true, "local": true}
        ],
        // Corresponds to 'n_jobs' of ZcurveEncoder.__init__ */
        "n_jobs": 8
    }
}

Tips: Use the Z-curve method more flexibly through our ZcurvePy's APIs! 
"""

# help information
arg_fasta = "    input sequence file as FASTA format (*.fa; *.fasta; *.fna)"
arg_gbk   = "    input sequence file as GenBank format (*.gb; *.genbank; *.gbk; *.gbff)"
arg_acc   = "    input sequence file as NCBI accession number (*.txt)"
arg_set   = "    hyper parameter setting file as json (*.json)"
arg_ext   = "    extract protein genes as samples from annotated sequence file"
arg_out   = "    output z-curve parameters file as comma-splited table (*.csv)"

def mergeInputs(
    fasta: str, 
    genbank: str, 
    accession: str, 
    extract: bool,
    command_name=__name__
):
    """
    Merge inputs from different source
    
    Args:
        fasta (str):        comma-splited paths to fasta files
        genbank (str):      comma-splited paths to genbank files
        accession (str):    comma-splited NCBI accession numbers
        extract (bool):     extract protein genes
        command_name (str): command_name
    """
    records = []
    try:
        # read fastas
        if fasta is not None:
            single_fastas = fasta.split(',')
            for single_fasta in single_fastas:
                records += list(parse(single_fasta, "fasta"))

        # read genbanks
        if genbank is not None:
            single_genbanks = genbank.split(',')
            for single_genbank in single_genbanks:
                if extract:
                    records += extract_CDS(single_genbank)
                else:
                    records += list(parse(single_genbank, "genbank"))
        
        # read accesions
        if accession is not None:
            form = 'gb' if extract else 'fasta'
            single_records = download_acc(accession, command_name, form)
            for single_record in single_records:
                if extract:
                    records += extract_CDS(single_record)
                else:
                    records += list(parse(single_record, "fasta"))
        
    except Exception as exception:
        print(f"Biopython: [error] failed to read input files! ")
        print(exception)
    
    if len(records) == 0:
        print(f"{command_name}: [error] none valid input provided! ")
        sys.exit(0)
    
    return records


def main():
    start_time = time.time()
    # program name (command name)
    command_name = 'zcurve-encoder'

    # create arguments parser and parse arguments
    parser = argparse.ArgumentParser(
        prog=command_name,
        description=prog_infor,
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=epilog_infor
    )
    parser.add_argument('-f', '--fasta', metavar='\b', required=False, type=str, default=None, help=arg_fasta)
    parser.add_argument('-g', '--genbank', metavar='\b', required=False, type=str, default=None, help=arg_gbk)
    parser.add_argument('-a', '--accession', metavar='\b', required=False, type=str, default=None, help=arg_acc)
    parser.add_argument('-e', '--extract', metavar='\b', required=False, default=False, type=bool, help=arg_ext)
    parser.add_argument('-s', '--setting', metavar='\b', required=True, type=str, help=arg_set)
    parser.add_argument('-o', '--output', metavar='\b', required=True, type=str, help=arg_out)
    args = vars(parser.parse_args())

    # extract/download sequences and merge inputs
    records = mergeInputs(
        fasta=args['fasta'], 
        genbank=args['genbank'],
        accession=args['accession'],
        extract=args['extract'],
        command_name=command_name
    )
    count = len(records)

    # load setting file and create encoder
    with open(args['setting'], 'r', encoding='utf-8') as file:
        setting = json.load(file)['encoder']

    if 'hyper_params' not in setting.keys():
        print(f"{command_name}: [error] Hyper-param settings are needed to complete the process. ")
        sys.exit()
    
    n_jobs = int(setting['n_jobs']) if 'n_jobs' in setting.keys() else -1
    encoder = BatchZcurveEncoder(hyper_params=setting['hyper_params'], n_jobs=n_jobs)

    # run coding process
    output_file = args['output']
    result = encoder.dump(records, output_file)
    used_time = time.time() - start_time

    if result:
        print(f"{command_name}: Successfully encoded {count} sequences in {round(used_time, 3)} s")
    else: 
        print(f"{command_name}: Unable to open {output_file}")



from ZcurvePy.Scripts.RunZcurvePlotter3D import *

def handleInputs2D(
    records: list, 
    args: dict, 
    command_name: str,
):
    with open(args['settings'], 'r', encoding='utf-8') as file:
        settings = json.load(file)['plotter']
    num_settings, curve_list = len(settings), []

    for i, record in enumerate(records):
        if i < num_settings:
            setting = settings[i]

            start, stop = 0, len(record)
            if 'start' in setting.keys():
                start = int(setting['start'])
            if 'stop' in setting.keys():
                stop = int(setting['stop'])
            
            if start > stop:
                record = record[start:] + record[:stop]
            elif start < stop:
                record = record[start:stop]
            else:
                print(f"{command_name}: [error] The slice length of {record.id} is zero!")
                sys.exit()

            if "comp" in setting.keys() and setting['comp']:
                record_id = record.id
                record = record.reverse_complement()
                record.id = record_id
            
            intv = int(setting['intv'] if "intv" in setting.keys() 
                       else max(len(record)/ 1E7, 1))
            curve_types = ['RY', 'MK', 'WS']

            if "curve2d" in setting.keys():
                curve_types = setting['curve2d'].split(",")
            else:
                print(f"{command_name}: [error] No valid settings found for 2D-mode !")
                sys.exit()
            
            curves = []
            plotter = ZcurvePlotter(record)

            for typ in curve_types:
                if typ not in types.keys():
                    print(f"{command_name}: [error] Invalid curve type {typ}")
                    sys.exit()
                window = int(setting['window']) if 'window' in setting.keys() else 0
                retrs = methodcaller(types[typ], window=window, return_n=False)(plotter)
                yvalues = retrs[0] if isinstance(retrs, tuple) else retrs
                curves.append((yvalues, typ))
                
            curve_list.append((record.id, curves, intv, len(record)))
    
    return curve_list


def visualize2D(curve_list: list, png: str, show: bool):
    plt.rc('font', family='Times New Roman')
    plt.rcParams.update({"font.size": 14})
    fig = plt.figure(figsize=(10, 6))

    ax = fig.add_subplot()
    ax.set_title('Z-curve')

    for item in curve_list:
        name, curves, intv, length = item
        n = np.arange(1, length + 1, intv)
        for y, typ in curves:
            ax.plot(n, y[::intv], label=f"{name} | {names[typ]}")
        
    ax.set_xlabel("n (bp)", labelpad=10)
    ax.set_ylabel("Component", labelpad=10)
    ax.legend()
    
    if png is not None:
        fig.savefig(png)
    if show and HAS_GUI:
        try:
            manager = plt.get_current_fig_manager()
            fig.canvas.manager.set_window_title("Z-curve Plotter (Powered by Matplotlib)")
            res_path = pkg_resources.resource_filename("ZcurvePy", "assets/image/icon.ico")
            manager.window.iconbitmap(res_path)
            plt.show()
        except Exception as exception:
            print("zcurve-plotter: [error] unable to create GUI! ")
            print(exception)


def main():
    command_name = 'zcurve-plotter'

    parser = argparse.ArgumentParser(
        prog=command_name,
        description=prog_infor,
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=epilog_infor
    )

    parser.add_argument('-f', '--fasta', metavar='\b', required=False, type=str, help=arg_fasta)
    parser.add_argument('-g', '--genbank', metavar='\b', required=False, type=str, help=arg_gbk)
    parser.add_argument('-a', '--accession', metavar='\b', required=False, type=str, help=arg_acc)
    parser.add_argument('-s', '--settings', metavar='\b', required=True, type=str, help=arg_set)
    parser.add_argument('-o', "--output", metavar='\b', required=False, default=None, type=str, help=arg_out)
    parser.add_argument('-p', "--png", metavar='\b', required=False, type=str, default=None, help=arg_png)
    parser.add_argument('-v', "--show", metavar='\b', required=False, type=bool, default=True, help=arg_show)

    args = vars(parser.parse_args())

    records = checkInputs(args, command_name)
    curve_list = handleInputs2D(records, args, command_name)

    png, show = args['png'], args['show']
    if png is not None or show:
        visualize2D(curve_list, png, show)
    
    output = args['output']
    if output is not None:
        contents = []
        for item in curve_list:
            name, curves, _, _ = item
            contents.append({name: {typ: np.round(curve, decimals=2).tolist() for curve, typ in curves}})
        to_save = json.dumps(contents)
        with open(output, "w") as file:
            file.write(to_save)


""" This is the main program of Z-curve Segmenter """
# PASS 2025-03-03
import sys
import pkg_resources
from ZcurvePy import ZcurveSegmenter, ZcurvePlotter, __version__
from Bio.SeqIO import parse
from ZcurvePy.Util import download_acc, has_gui
import matplotlib
HAS_GUI = has_gui()
matplotlib.use("TkAgg" if HAS_GUI else "Agg")
from matplotlib import pyplot as plt
from _ZcurvePy import ZcurvePlotter
from operator import methodcaller
import argparse

# program information
prog_infor = \
f"""
Z-curve Segmenter -- A program to segment Z-curves

Authors:    Zhang Zetong, Gao Feng
Version:    {__version__}
Contact:    fgao@tju.edu.cn
Copyright:  Bioinformatic Center, Tianjin University, China
"""

# appendix information
epilog_infor = \
"""
Modes:
(1) GN: segmentation for Genome        (S(P) = a^2 + t^2 + g^2 + c^2)
(2) RY: segmentation for RY-disparity  (S(P) = (a + t)^2 + (g + c)^2)
(3) MK: segmentation for MK-disparity  (S(P) = (a + c)^2 + (g + t)^2)
(4) WS: segmentation for WS-disparity  (S(P) = (a + t)^2 + (g + c)^2)
(5) AT: segmentation for AT-disparity  (S(P) = a^2 + t^2)
(6) GC: segmentation for GC-disparity  (S(P) = g^2 + c^2)
(7) CG: segmentation for CpG-profile   (S(P) = CpG^2 + (1-CpG)^2)

Tips: Use the Z-curve method more flexibly through our ZcurvePy's APIs! 
"""

# help information
arg_fasta = "    input sequence file as FASTA format (*.fa; *.fasta; *.fna)"
arg_gbk   = "    input sequence file as GenBank format (*.gb; *.genbank; *.gbk; *.gbff)"
arg_acc   = "    input sequence file as NCBI accession number (*.txt)"
arg_start = "    start point of the subsequence to be segmented. (Default: 0)"
arg_end   = "    end point of the subsequence to be segmented. (Default: [Sequence Length])"
arg_mode  = "    choose the mode of Z-curve Segmenter based on S(P) (Default: 'GN')"
arg_halt  = "    halting parameter for the segmenting recursion (Default: 100)"
arg_min   = "    the min length between two segmentation point (Default: 3000 bp)"
arg_depth = "    the max depth of the segmenting recursion (Default: 9999)"
arg_out   = "    output segmentation points as comma-splited table (*.csv)"
arg_png   = "    output graphic files as PNG picture (*.png)"
arg_show  = "    show the curve and segmentation points (Default: False)"

# curve types
types = {
    "GN"  : [("RY_disparity", "RY-disparity"), 
             ("MK_disparity", "MK-disparity"), 
             ("z_prime_curve", "z' curve")],
    "RY"  : [("x_prime_curve", "x' curve")],
    "MK"  : [("y_prime_curve", "y' curve")],
    "WS"  : [("z_prime_curve", "z' curve")],
    "AT"  : [("AT_prime_curve", "d'(AT) curve")],
    "GC"  : [("GC_prime_curve", "d'(GC) curve")],
    "CG"  : [("CpG_prime_curve", "CpG-profile")],
}

def handleInputs(
    fasta: str, 
    genbank: str, 
    accession: str, 
    command_name=__name__
):
    """
    Handle inputs from different source
    
    Args:
        fasta (str):        comma-splited paths to fasta files
        genbank (str):      comma-splited paths to genbank files
        accession (str):    comma-splited NCBI accession numbers
        extract (bool):     extract protein genes
        command_name (str): command_name
    """
    record = None
    try:
        if fasta is not None:
            record = next(parse(fasta, "fasta"))
        elif genbank is not None:
            record = next(parse(genbank, "gb"))
        elif accession is not None:
            ready = download_acc(accession, command_name)
            record = next(parse(ready[0], "fasta"))
    except Exception as exception:
        print(f"Biopython: [error] failed to read input files! ")
        print(exception)
    
    if record is not None:
        return record
    else:
        print(f"{command_name}: [error] no valid input was found! ")
        sys.exit(0)


def visualize(
    record,
    mode,
    plot_range,
    seg_points,
    show=False,
    save=None
):
    intv = 10
    plt.rc('font', family='Times New Roman')
    plt.rcParams.update({"font.size": 14})
    plotter = ZcurvePlotter(record)
    fig = plt.figure(figsize=(10, 6))
    ax = fig.add_subplot()
    ax.set_title(f'Mode = {mode}')
    ax.set_xlabel("n (bp)", labelpad=10)
    ax.set_ylabel("Component", labelpad=10)

    for typ, name in types[mode]:
        results = methodcaller(typ, return_n=False)(plotter)
        if "prime" in typ:
            results = results[0]
        ax.plot(plot_range[::intv], results[::intv], label=name)
    for point, _ in seg_points:
        ax.axvline(point, color='red')
    ax.legend()
    if save is not None:
        fig.savefig(save)
    if show and HAS_GUI:
        try:
            manager = plt.get_current_fig_manager()
            fig.canvas.manager.set_window_title("Z-curve Plotter (Powered by Matplotlib)")
            res_path = pkg_resources.resource_filename("ZcurvePy", "assets/image/icon.ico")
            manager.window.iconbitmap(res_path)
            plt.show()
        except Exception as exception:
            print("zcurve-plotter: [error] unable to create GUI! ")
            print(exception)

def main():
    # program name (command name)
    command_name = 'zcurve-segmenter'

    # create arguments parser and parse arguments
    parser = argparse.ArgumentParser(
        prog=command_name,
        description=prog_infor,
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=epilog_infor
    )
    parser.add_argument('-f', '--fasta', metavar='\b', required=False, type=str, default=None, help=arg_fasta)
    parser.add_argument('-g', '--genbank', metavar='\b', required=False, type=str, default=None, help=arg_gbk)
    parser.add_argument('-a', '--accession', metavar='\b', required=False, type=str, default=None, help=arg_acc)
    parser.add_argument('-s', '--start', metavar='\b', required=False, type=int, default=0, help=arg_start)
    parser.add_argument('-e', '--end', metavar='\b', required=False, type=int, default=-1, help=arg_end)
    parser.add_argument('-m', '--mode', metavar='\b', required=True, type=str, default='GN', help=arg_mode)
    parser.add_argument('-t', '--halting', metavar='\b', required=False, type=float, default=100, help=arg_halt)
    parser.add_argument('-l', '--minlen', metavar='\b', required=False, type=int, default=3000, help=arg_min)
    parser.add_argument('-d', '--maxdepth', metavar='\b', required=False, type=int, default=9999, help=arg_depth)
    parser.add_argument('-o', '--output', metavar='\b', required=True, type=str, help=arg_out)
    parser.add_argument('-p', '--png', metavar='\b', required=False, default=None, type=str, help=arg_png)
    parser.add_argument('-v', '--show', metavar='\b', required=False, type=bool, default=False, help=arg_show)
    args = vars(parser.parse_args())

    record = handleInputs(args['fasta'], args['genbank'], args['accession'])
    length, start, end = len(record), args['start'], args['end']
    end = length if end < 0 else end
    
    mode, halting, minlen, maxdepth = args['mode'], int(args['halting']), int(args['minlen']), args['maxdepth']

    plot_range = list(range(start - length, 0)) + list(range(0, end)) if start > end else list(range(start, end))
    record = record[start:] + record[:end] if start > end else record[start:end]
    
    if mode not in ZcurveSegmenter.func.keys():
        print(f"{command_name}: [error] unknown mode '{mode}'. ")
        sys.exit()
    
    segmenter = ZcurveSegmenter(mode=mode, halting=halting, min_len=minlen, max_depth=maxdepth)
    seg_points = segmenter.run(record)
    real_locs = [[plot_range[i], v] for i, v in seg_points]

    with open(args['output'], "w") as output:
        output.write("No,Seg_Point,dS_Value\n")
        for i, (point, value) in enumerate(real_locs):
            output.write(f"{i + 1},{point},{round(value, 2)}\n")
    
    png, show = args['png'], args['show']

    if png is not None or show:
        visualize(record, mode, plot_range, real_locs, show, png)
    
    print(f"Successfully segmented at {len(seg_points)} points on the genome.")



from setuptools import Extension, setup, find_packages
import numpy as np

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("LICENSE", "r", encoding="utf-8") as fh:
    license = fh.read()

with open("requirements.txt", encoding="utf-8") as fh:
    requirements = fh.readlines()

setup(
    name='zcurvepy',
    author='TUBIC',
    author_email='fgao@tju.edu.cn',
    maintainer='Zetong Zhang',
    maintainer_email='zhangzetong@tju.edu.cn',
    url='https://zcurvehub-docs.readthedocs.io/',
    download_url='https://pypi.org/project/zcurvepy/',
    description='High performance Python toolkit for the Z-curve theory',
    # long_description=long_description,
    # long_description_content_type="text/markdown",
    keywords=['Bioinformatics', 'Z-curve', 'Machine Learning'],
    license=license,
    version='1.6.0',
    install_requires=requirements,
    python_requires='>=3.7, <=3.11',
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={"ZcurvePy.assets.image": ["*.png", "*.jpg", "*.ico"]},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'zcurve-encoder=ZcurvePy.Scripts.RunZcurveEncoder:main',
            'zcurve-plotter=ZcurvePy.Scripts.RunZcurvePlotter:main',
            'zcurve-plotter-3d=ZcurvePy.Scripts.RunZcurvePlotter3D:main',
            'zcurve-segmenter=ZcurvePy.Scripts.RunZcurveSegmenter:main',
        ],
    },
    ext_modules=[Extension("_ZcurvePy", 
        include_dirs=[np.get_include()],
        sources=[
            "ZcurvePy/include/ZcurvePy.cpp", 
            "ZcurvePy/include/ZcurvePyCore.cpp", 
            "ZcurvePy/include/ZcurvePyAPIs.cpp",
            "ZcurvePy/include/ZislandFinder.cpp"
        ]
    )]
)
